import asyncio
import logging
import collections
from functools import partial

from .constants import BUFSIZE
from .utils import wall_clock_sleep


class InappropriateRead(Exception):
    pass


class SuccessError(Exception):
    pass


class TTLExpired(Exception):
    pass


class ConnPool:
    def __init__(self, *,
                 dst_address,
                 dst_port,
                 ssl_context=True,
                 ssl_hostname=None,
                 timeout=5,
                 backoff=5,
                 ttl=30,
                 size=10,
                 loop=None):
        self._logger = logging.getLogger(self.__class__.__name__)
        self._loop = loop if loop is not None else asyncio.get_event_loop()
        self._dst_address = dst_address
        self._dst_port = dst_port
        self._ssl_context = ssl_context
        self._ssl_hostname = ssl_hostname
        self._timeout = timeout
        self._ttl = ttl
        self._size = size
        self._backoff = backoff
        self._waiters = collections.deque()
        self._reserve = collections.deque()
        self._conn_builders = set()

    async def start(self):
        self._conn_builders = set(self._loop.create_task(self._build_conn())
                                  for _ in range(self._size))

    async def stop(self):
        while self._conn_builders:
            tasks = list(self._conn_builders)
            self._conn_builders.clear()
            for t in tasks:
                t.cancel()
            await asyncio.gather(*tasks, return_exceptions=True)
        for (reader, writer), _ in self._reserve:
            writer.close()

    async def _build_conn(self):
        async def fail():
            self._logger.debug("Failed upstream connection. Backoff for %d "
                               "seconds", self._backoff)
            await wall_clock_sleep(self._backoff)

        async def fail_corrupted():
            self._logger.warning("Upstream connection corrupted. Backoff for"
                                 " %d seconds", self._backoff)
            await wall_clock_sleep(self._backoff)

        async def reader_guard(reader):
            try:
                await reader.read(1)
            except asyncio.CancelledError:
                raise
            except Exception as exc:
                self._logger.debug("reader_guard catched exception on "
                                   "reader.read(): %s", str(exc))
            raise InappropriateRead()

        async def waiter_guard(event):
            await event.wait()
            raise SuccessError()

        async def timeout_guard(timeout):
            await wall_clock_sleep(timeout)
            raise TTLExpired()

        async def taker(grabbed, read_task):
            grabbed.set()
            if read_task.done():
                raise InapproptiateRead()
            else:
                try:
                    await read_task
                except asyncio.CancelledError:
                    pass
                else:
                    raise InappropriateRead()

        def try_remove_from_queue(elem):
            try:
                self._reserve.remove(elem)
            except ValueError:
                self._logger.debug("Not found expired connection "
                                   "in reserve. This should not happen.")
            else:
                elem[0][1].close()

        while True:
            try:
                try:
                    conn = await asyncio.wait_for(
                        asyncio.open_connection(self._dst_address,
                                                self._dst_port,
                                                ssl=self._ssl_context,
                                                server_hostname=self._ssl_hostname),
                        self._timeout)
                except asyncio.TimeoutError:
                    self._logger.error("Connection to upstream timed out.")
                    await fail()
                except asyncio.CancelledError:
                    raise
                except Exception as exc:
                    self._logger.exception("Got exception while connecting to upstream: %s", str(exc))
                    await fail()
                else:
                    self._logger.debug("Successfully built upstream connection.")
                    while self._waiters:
                        fut = self._waiters.popleft()
                        if not fut.cancelled():
                            self._logger.warning("Pool exhausted. Dispatching connection directly to waiter!")
                            fut.set_result(conn)
                            break
                    else:
                        grabbed = asyncio.Event()

                        read_task = asyncio.ensure_future(reader_guard(conn[0]))
                        tasks = [
                            asyncio.ensure_future(waiter_guard(grabbed)),
                            asyncio.ensure_future(timeout_guard(self._ttl)),
                            read_task,
                        ]

                        elem = (conn, partial(taker, grabbed, read_task))
                        self._reserve.append(elem)

                        try:
                            await asyncio.gather(*tasks)
                        except InappropriateRead:
                            if not grabbed.is_set():
                                try_remove_from_queue(elem)
                            await fail_corrupted()
                        except TTLExpired:
                            if not grabbed.is_set():
                                try_remove_from_queue(elem)
                        except SuccessError:
                            pass
                        finally:
                            for task in tasks:
                                if not task.done():
                                    task.cancel()
            except asyncio.CancelledError:
                raise
            except Exception as exc:
                self._logger.exception("_build_conn crashed with exception: %s",
                                       str(exc))
                    
    async def get(self):
        while True:
            if self._reserve:
                conn, take = self._reserve.popleft()
                try:
                    await take()
                except InappropriateRead:
                    pass
                else:
                    self._logger.debug("Obtained connection from pool.")
                    return conn
            else:
                fut = self._loop.create_future()
                self._waiters.append(fut)
                self._logger.debug("Awaiting for free connection.")
                return await fut



import sys
import argparse
import asyncio
import logging
import ssl
import signal
from functools import partial
from urllib.parse import urlparse

from .asdnotify import AsyncSystemdNotifier

from .listener import Listener
from .constants import LogLevel
from .proxy_protocol import ProxyProtocol, check_proxyprotocol
from . import utils
from .connpool import ConnPool


def parse_args():
    parser = argparse.ArgumentParser(
        description="Pooling TLS wrapper",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument("dst_address",
                        help="target hostname")
    parser.add_argument("dst_port",
                        type=utils.check_port,
                        help="target port")
    parser.add_argument("-v", "--verbosity",
                        help="logging verbosity",
                        type=utils.check_loglevel,
                        choices=LogLevel,
                        default=LogLevel.info)
    parser.add_argument("-l", "--logfile",
                        help="log file location",
                        metavar="FILE")
    parser.add_argument("--disable-uvloop",
                        help="do not use uvloop even if it is available",
                        action="store_true")

    listen_group = parser.add_argument_group('listen options')
    listen_group.add_argument("-a", "--bind-address",
                              default="127.0.0.1",
                              help="bind address")
    listen_group.add_argument("-p", "--bind-port",
                              default=57800,
                              type=utils.check_port,
                              help="bind port")
    listen_group.add_argument("-W", "--pool-wait-timeout",
                              default=15,
                              type=utils.check_positive_float,
                              help="timeout for pool await state of client "
                              "connection")
    listen_group.add_argument("-P", "--proxy-protocol",
                              default=ProxyProtocol.none,
                              choices=ProxyProtocol,
                              type=check_proxyprotocol,
                              help="transparent mode: prepend all connections"
                              " with proxy-protocol data")

    pool_group = parser.add_argument_group('pool options')
    pool_group.add_argument("-n", "--pool-size",
                            default=25,
                            type=utils.check_positive_int,
                            help="connection pool size")
    pool_group.add_argument("-B", "--backoff",
                            default=5,
                            type=utils.check_positive_float,
                            help="delay after connection attempt failure in seconds")
    pool_group.add_argument("-T", "--ttl",
                            default=30,
                            type=utils.check_positive_float,
                            help="lifetime of idle pool connection in seconds")
    pool_group.add_argument("-w", "--timeout",
                            default=4,
                            type=utils.check_positive_float,
                            help="server connect timeout")

    tls_group = parser.add_argument_group('TLS options')
    tls_group.add_argument("-c", "--cert",
                           help="use certificate for client TLS auth")
    tls_group.add_argument("-k", "--key",
                           help="key for TLS certificate")
    tls_group.add_argument("-C", "--cafile",
                           help="override default CA certs "
                           "by set specified in file")
    ssl_name_group=tls_group.add_mutually_exclusive_group()
    ssl_name_group.add_argument("--no-hostname-check",
                                action="store_true",
                                help="do not check hostname in cert subject. "
                                "This option is useful for private PKI and "
                                "available only together with \"--cafile\"")
    ssl_name_group.add_argument("--tls-servername",
                                type=utils.check_ssl_hostname,
                                help="specifies hostname to expect in server "
                                "TLS certificate")
    return parser.parse_args()


async def amain(args, loop):  # pragma: no cover
    logger = logging.getLogger('MAIN')

    context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
    ssl_hostname = None
    if args.cafile:
        context.load_verify_locations(cafile=args.cafile)
    if args.no_hostname_check:
        if not args.cafile:
            logger.fatal("CAfile option is required when hostname check "
                         "is disabled. Terminating program.")
            sys.exit(2)
        ssl_hostname = ''
    elif args.tls_servername:
        ssl_hostname = args.tls_servername
    if args.cert:
        context.load_cert_chain(certfile=args.cert, keyfile=args.key)


    proxy_protocol = args.proxy_protocol.value() if args.proxy_protocol.value else None
    pool = ConnPool(dst_address=args.dst_address,
                    dst_port=args.dst_port,
                    ssl_context=context,
                    ssl_hostname=ssl_hostname,
                    timeout=args.timeout,
                    backoff=args.backoff,
                    ttl=args.ttl,
                    size=args.pool_size,
                    loop=loop)
    await pool.start()
    server = Listener(listen_address=args.bind_address,
                      listen_port=args.bind_port,
                      timeout=args.pool_wait_timeout,
                      pool=pool,
                      proxy_protocol=proxy_protocol,
                      loop=loop)
    await server.start()
    logger.info("Server started.")

    exit_event = asyncio.Event()
    async with utils.Heartbeat():
        sig_handler = partial(utils.exit_handler, exit_event)
        signal.signal(signal.SIGTERM, sig_handler)
        signal.signal(signal.SIGINT, sig_handler)
        async with AsyncSystemdNotifier() as notifier:
            await notifier.notify(b"READY=1")
            await exit_event.wait()

            logger.debug("Eventloop interrupted. Shutting down server...")
            await notifier.notify(b"STOPPING=1")
    await server.stop()
    await pool.stop()


def main():  # pragma: no cover
    args = parse_args()
    with utils.AsyncLoggingHandler(args.logfile) as log_handler:
        logger = utils.setup_logger('MAIN', args.verbosity, log_handler)
        utils.setup_logger('Listener', args.verbosity, log_handler)
        utils.setup_logger('ConnPool', args.verbosity, log_handler)

        logger.info("Starting eventloop...")
        if not args.disable_uvloop:
            if utils.enable_uvloop():
                logger.info("uvloop enabled.")
            else:
                logger.info("uvloop is not available. "
                            "Falling back to built-in event loop.")

        loop = asyncio.get_event_loop()
        # workaround for Python bug on pending writes to SSL connections
        utils.ignore_ssl_error(loop)
        loop.run_until_complete(amain(args, loop))
        loop.close()
        logger.info("Server finished its work.")



from abc import ABC, abstractmethod
import string
import enum
import socket
import struct

from .utils import detect_af


af_map = {
    socket.AF_INET: "TCP4",
    socket.AF_INET6: "TCP6",
}


class BaseProxyProtocol(ABC):
    @abstractmethod
    def prologue(self, src, dst):
        """ Returns bytes with prologue data. Accepts original
        source and destination of TCP connection. 
        Params: src, dst
        Each of them is a pair of host (IP as str) and port (int) """


class ProxyProtocolV1(BaseProxyProtocol):
    def prologue(self, src, dst):
        src_ip, src_port = src
        dst_ip, dst_port = dst
        src_af = detect_af(src_ip)
        dst_af = detect_af(src_ip)
        if src_af != dst_af:
            raise ValueError("Protocol AF mismatch!")
        if src_af not in af_map:
            return b"PROXY UNKNOWN\r\n"
        else:
            res= ("PROXY " + af_map[src_af] + " " + 
                  src_ip + " " + dst_ip + " " +
                  str(src_port) + " " + str(dst_port) + "\r\n").encode('ascii')
            if len(res) >= 108:
                raise RuntimeError("Produced string is too long for proxy-protocol")
            return res


PPV2SIG = b'\x0D\x0A\x0D\x0A\x00\x0D\x0A\x51\x55\x49\x54\x0A'
PPV2StructIPv4 = struct.Struct("!12sBBH" "4s4sHH")
PPV2StructIPv6 = struct.Struct("!12sBBH" "16s16sHH")
PPV2StructUNKNOWN = struct.Struct("!12sBBH")
PPV2VER = 2
PPV2PROXYCMD = 1
PPV2VERCMD = ((PPV2VER << 4) | PPV2PROXYCMD)
PPV2TCPOVERIPV4 = 0x11
PPV2TCPOVERIPV6 = 0x21
PPV2UNKNOWNAF = 0x0
PPV2UNKNOWNAF_ADDRLEN = 0x0
PPV2TCPOVERIPV4_ADDRLEN = 12
PPV2TCPOVERIPV6_ADDRLEN = 36


class ProxyProtocolV2(BaseProxyProtocol):
    def prologue(self, src, dst):
        src_ip, src_port = src
        dst_ip, dst_port = dst
        src_af = detect_af(src_ip)
        dst_af = detect_af(src_ip)
        if src_af != dst_af:
            raise ValueError("Protocol AF mismatch!")
        src_addr_packed = socket.inet_pton(src_af, src_ip)
        dst_addr_packed = socket.inet_pton(dst_af, dst_ip)
        if src_af == socket.AF_INET:
            # IPv4
            res = PPV2StructIPv4.pack(PPV2SIG, PPV2VERCMD, PPV2TCPOVERIPV4,
                                      PPV2TCPOVERIPV4_ADDRLEN,
                                      src_addr_packed,
                                      dst_addr_packed,
                                      src_port,
                                      dst_port)
        elif src_af == socket.AF_INET6:
            # IPv6
            res = PPV2StructIPv6.pack(PPV2SIG, PPV2VERCMD, PPV2TCPOVERIPV6,
                                      PPV2TCPOVERIPV6_ADDRLEN,
                                      src_addr_packed,
                                      dst_addr_packed,
                                      src_port,
                                      dst_port)
        else:
            res = PPV2StructUNKNOWN.pack(PPV2SIG, PPV2VERCMD, PPV2UNKNOWNAF,
                                      PPV2UNKNOWNAF_ADDRLEN)
        return res


class ProxyProtocol(enum.Enum):
    none = None
    v1 = ProxyProtocolV1
    v2 = ProxyProtocolV2

    def __str__(self):
        return self.name


def check_proxyprotocol(arg):
    try:
        return ProxyProtocol[arg]
    except (IndexError, KeyError):
        raise argparse.ArgumentTypeError("%s is not valid proxy-protocol" % (repr(arg),))



import asyncio
import argparse
import logging
import logging.handlers
import ssl
import os
import queue
import socket
import ctypes
import time

from . import constants


def ignore_ssl_error(loop):
    """Ignore aiohttp #3535 issue with SSL data after close

    There appears to be an issue on Python 3.7 and aiohttp SSL that throws a
    ssl.SSLError fatal error (ssl.SSLError: [SSL: KRB5_S_INIT] application data
    after close notify (_ssl.c:2609)) after we are already done with the
    connection. See GitHub issue aio-libs/aiohttp#3535

    Given a loop, this sets up a exception handler that ignores this specific
    exception, but passes everything else on to the previous exception handler
    this one replaces.

    """
    orig_handler = loop.get_exception_handler()

    def ignore_ssl_error(loop, context):
        if context.get('message') == 'SSL error in data received':
            # validate we have the right exception, transport and protocol
            exception = context.get('exception')
            protocol = context.get('protocol')
            if (
                isinstance(exception, ssl.SSLError) and exception.reason == 'KRB5_S_INIT' and
                isinstance(protocol, asyncio.sslproto.SSLProtocol) and
                'application data after close notify' in exception.strerror
            ):
                if loop.get_debug():
                    asyncio.log.logger.debug('Ignoring aiohttp SSL KRB5_S_INIT error')
                return
        if orig_handler is not None:
            orig_handler(loop, context)
        else:
            loop.default_exception_handler(context)

    loop.set_exception_handler(ignore_ssl_error)


class OverflowingQueue(queue.Queue):
    def put(self, item, block=True, timeout=None):
        try:
            return queue.Queue.put(self, item, block, timeout)
        except queue.Full:
            pass

    def put_nowait(self, item):
        return self.put(item, False)


class AsyncLoggingHandler:
    def __init__(self, logfile=None, maxsize=1024):
        _queue = OverflowingQueue(maxsize)
        if logfile is None:
            _handler = logging.StreamHandler()
        else:
            _handler = logging.FileHandler(logfile)
        self._listener = logging.handlers.QueueListener(_queue, _handler)
        self._async_handler = logging.handlers.QueueHandler(_queue)

        _handler.setFormatter(logging.Formatter('%(asctime)s '
                                                '%(levelname)-8s '
                                                '%(name)s: %(message)s',
                                                '%Y-%m-%d %H:%M:%S'))

    def __enter__(self):
        self._listener.start()
        return self._async_handler

    def __exit__(self, exc_type, exc_value, traceback):
        self._listener.stop()


def setup_logger(name, verbosity, handler):
    logger = logging.getLogger(name)
    logger.setLevel(verbosity)
    logger.addHandler(handler)
    return logger


def check_port(value):
    def fail():
        raise argparse.ArgumentTypeError(
            "%s is not a valid port number" % value)
    try:
        ivalue = int(value)
    except ValueError:
        fail()
    if not 0 < ivalue < 65536:
        fail()
    return ivalue


def check_positive_float(value):
    def fail():
        raise argparse.ArgumentTypeError(
            "%s is not a valid value" % value)
    try:
        fvalue = float(value)
    except ValueError:
        fail()
    if fvalue <= 0:
        fail()
    return fvalue


def check_positive_int(value):
    def fail():
        raise argparse.ArgumentTypeError(
            "%s is not a valid value" % value)
    try:
        fvalue = int(value)
    except ValueError:
        fail()
    if fvalue <= 0:
        fail()
    return fvalue


def check_loglevel(arg):
    try:
        return constants.LogLevel[arg]
    except (IndexError, KeyError):
        raise argparse.ArgumentTypeError("%s is not valid loglevel" % (repr(arg),))


def check_ssl_hostname(arg):
    if not arg:
        raise argparse.ArgumentTypeError("%s is not valid server name" % (repr(arg),))
    return arg


def enable_uvloop():  # pragma: no cover
    try:
        import uvloop
        asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
    except ImportError:
        return False
    else:
        return True


def exit_handler(exit_event, signum, frame):  # pragma: no cover pylint: disable=unused-argument
    logger = logging.getLogger('MAIN')
    if exit_event.is_set():
        logger.warning("Got second exit signal! Terminating hard.")
        os._exit(1)  # pylint: disable=protected-access
    else:
        logger.warning("Got first exit signal! Terminating gracefully.")
        exit_event.set()


class Heartbeat:
    def __init__(self, interval=.5):
        self._interval = interval
        self._beat = None

    async def heartbeat(self):
        while True:
            await asyncio.sleep(self._interval)

    async def __aenter__(self):
        return await self.start()

    async def start(self):
        if self._beat is None:
            self._beat = asyncio.ensure_future(self.heartbeat())
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        return await self.stop()

    async def stop(self):
        self._beat.cancel()
        while not self._beat.done():
            try:
                await self._beat
            except asyncio.CancelledError:
                pass


def detect_af(addr):
    return socket.getaddrinfo(addr,
                              None,
                              socket.AF_UNSPEC,
                              0,
                              0,
                              socket.AI_NUMERICHOST)[0][0]

class sockaddr(ctypes.Structure):
    _fields_ = [('sa_family', ctypes.c_uint16),
                ('sa_data', ctypes.c_char * 14),
               ]


class sockaddr_in(ctypes.Structure):
    _fields_ = [('sin_family', ctypes.c_uint16),
                ('sin_port', ctypes.c_uint16),
                ('sin_addr', ctypes.c_uint32),
               ]

sockaddr_size = max(ctypes.sizeof(sockaddr_in), ctypes.sizeof(sockaddr))


class sockaddr_in6(ctypes.Structure):
    _fields_ = [('sin6_family', ctypes.c_uint16),
                ('sin6_port', ctypes.c_uint16),
                ('sin6_flowinfo', ctypes.c_uint32),
                ('sin6_addr', ctypes.c_char * 16),
                ('sin6_scope_id', ctypes.c_uint32),
               ]

sockaddr6_size = ctypes.sizeof(sockaddr_in6)

def get_orig_dst(sock):
    own_addr = sock.getsockname()[0]
    own_af = detect_af(own_addr)
    if own_af == socket.AF_INET:
        buf = sock.getsockopt(socket.SOL_IP, constants.SO_ORIGINAL_DST, sockaddr_size)
        sa = sockaddr_in.from_buffer_copy(buf)
        addr = socket.ntohl(sa.sin_addr)
        addr = str(addr >> 24) + '.' + str((addr >> 16) & 0xFF) + '.' + str((addr >> 8) & 0xFF) + '.' + str(addr & 0xFF)
        port = socket.ntohs(sa.sin_port)
        return addr, port
    elif own_af == socket.AF_INET6:
        buf = sock.getsockopt(constants.SOL_IPV6, constants.SO_ORIGINAL_DST, sockaddr6_size)
        sa = sockaddr_in6.from_buffer_copy(buf)
        addr = socket.inet_ntop(socket.AF_INET6, sa.sin6_addr)
        port = socket.ntohs(sa.sin_port)
        return addr, port
    else:
        raise RuntimeError("Unknown address family!")

async def wall_clock_sleep(duration, precision=.2):
    async def _wall_clock_sleep():
        end_time = time.time() + duration
        while time.time() < end_time:
            await asyncio.sleep(precision)

    try:
        await asyncio.wait_for(_wall_clock_sleep(), duration)
    except asyncio.TimeoutError:
        pass



import asyncio
import logging
import collections
from functools import partial

from .constants import BUFSIZE
from .utils import get_orig_dst


class Listener:  # pylint: disable=too-many-instance-attributes
    def __init__(self, *,
                 listen_address,
                 listen_port,
                 pool,
                 timeout=None,
                 proxy_protocol=None,
                 loop=None):
        self._loop = loop if loop is not None else asyncio.get_event_loop()
        self._logger = logging.getLogger(self.__class__.__name__)
        self._listen_address = listen_address
        self._listen_port = listen_port
        self._children = set()
        self._server = None
        self._timeout = timeout
        self._conn_pool = pool
        self._proxy_protocol = proxy_protocol

    async def stop(self):
        self._server.close()
        await self._server.wait_closed()
        while self._children:
            children = list(self._children)
            self._children.clear()
            self._logger.debug("Cancelling %d client handlers...",
                               len(children))
            for task in children:
                task.cancel()
            await asyncio.wait(children)
            # workaround for TCP server keeps spawning handlers for a while
            # after wait_closed() completed
            await asyncio.sleep(.5)

    async def _pump(self, writer, reader):
        while True:
            data = await reader.read(BUFSIZE)
            if not data:
                break
            writer.write(data)
            await writer.drain()

    async def handler(self, reader, writer):
        peer_addr = writer.transport.get_extra_info('peername')
        self._logger.info("Client %s connected", str(peer_addr))
        if self._proxy_protocol:
            try:
                sock = writer.transport.get_extra_info('socket')
                orig_dst = get_orig_dst(sock)
                prologue = self._proxy_protocol.prologue(peer_addr, orig_dst)
                self._logger.debug("Client %s orig_dst=%s", str(peer_addr), str(orig_dst))
                self._logger.debug("Client %s prologue=%s", str(peer_addr), repr(prologue))
            except Exception as exc:
                self._logger.exception("Unable to handle connection transparency: "
                                   "%s", str(exc))
                return
        dst_writer = None
        try:
            dst_reader, dst_writer = await asyncio.wait_for(self._conn_pool.get(),
                                                            self._timeout)
            if self._proxy_protocol:
                dst_writer.write(prologue)
            t1 = asyncio.ensure_future(self._pump(writer, dst_reader))
            t2 = asyncio.ensure_future(self._pump(dst_writer, reader))
            try:
                await asyncio.gather(t1, t2)
            finally:
                for t in (t1, t2):
                    if not t.done():
                        t.cancel()
                        while not t.done():
                            try:
                                await t
                            except asyncio.CancelledError:
                                pass
        except asyncio.CancelledError:  # pylint: disable=try-except-raise
            raise
        except asyncio.TimeoutError:
            self._logger.error("Dropping client %s due to upstream connection "
                               "wait timed out.", peer_addr)
        except ConnectionResetError:
            self._logger.debug("Dropping client %s due to connection reset.", peer_addr)
        except Exception as exc:  # pragma: no cover
            self._logger.exception("Connection handler stopped with exception:"
                                   " %s", str(exc))
        finally:
            self._logger.info("Client %s disconnected", str(peer_addr))
            if dst_writer is not None:
                dst_writer.close()
            writer.close()

    async def start(self):
        def _spawn(reader, writer):
            def task_cb(task, fut):
                self._children.discard(task)
            task = self._loop.create_task(self.handler(reader, writer))
            self._children.add(task)
            task.add_done_callback(partial(task_cb, task))

        self._server = await asyncio.start_server(_spawn,
                                                  self._listen_address,
                                                  self._listen_port)
        self._logger.info("Server ready.")



from setuptools import setup

from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='ptw',
      version='1.0.1',
      description='Pooling TLS Wrapper',
      url='https://github.com/Snawoot/ptw',
      author='Vladislav Yarmak',
      author_email='vladislav-ex-src@vm-0.com',
      license='MIT',
      packages=['ptw'],
      python_requires='>=3.5.3',
      setup_requires=[
          'wheel',
      ],
      extras_require={
          'uvloop': 'uvloop>=0.11.0',
      },
      entry_points={
          'console_scripts': [
              'ptw=ptw.__main__:main',
          ],
      },
      classifiers=[
          "Programming Language :: Python :: 3.5",
          "License :: OSI Approved :: MIT License",
          "Operating System :: OS Independent",
          "Development Status :: 5 - Production/Stable",
          "Environment :: No Input/Output (Daemon)",
          "Intended Audience :: System Administrators",
          "Natural Language :: English",
          "Topic :: Internet",
          "Topic :: Security",
      ],
      long_description=long_description,
      long_description_content_type='text/markdown',
      zip_safe=True)



import os
import socket
import asyncio

MAX_QLEN = 128

class AsyncSystemdNotifier:
    def __init__(self):
        env_var = os.getenv('NOTIFY_SOCKET')
        self._addr = ('\0' + env_var[1:]
                      if env_var is not None and env_var.startswith('@')
                      else env_var)
        self._sock = None
        self._started = False
        self._loop = None
        self._queue = asyncio.Queue(MAX_QLEN)
        self._monitor = False

    @property
    def started(self):
        return self._started

    def _drain(self):
        while not self._queue.empty():
            msg = self._queue.get_nowait()
            self._queue.task_done()
            try:
                self._send(msg)
            except BlockingIOError:  # pragma: no cover
                self._monitor = True
                self._loop.add_writer(self._sock.fileno(), self._drain)
                break
            except OSError:
                pass
        else:
            if self._monitor:
                self._monitor = False
                self._loop.remove_writer(self._sock.fileno())

    def _send(self, data):
        return self._sock.sendto(data, socket.MSG_NOSIGNAL, self._addr)

    async def start(self):
        if self._addr is None:
            return False
        self._loop = asyncio.get_event_loop()
        try:
            self._sock = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
            self._sock.setblocking(0)
            self._started = True
        except OSError:
            return False
        return True

    async def notify(self, status):
        if self._started:
            await self._queue.put(status)
            self._drain()

    async def stop(self):
        if self._started:
            self._started = False
            await self._queue.join()
            if self._monitor:
                self._loop.remove_writer(self._sock.fileno())
            self._sock.close()

    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc, traceback):
        await self.stop()
